// Copyright 2018 The Exonum Team
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

//! Serialize structure into specific format.
//! Currently support only json.
//! This module is a pack of superstructures over serde `Serializer`\`Deserializer`

pub use hex::{decode as decode_hex, encode as encode_hex, FromHex, FromHexError, ToHex};

use super::Offset;
use encoding::Field;
use messages::MessageWriter;

/// Implements Exonum serialization / deserialization based on Serde `Serialize` / `Deserialize`.
///
/// The type supplied as the argument to this macro should implement the following
/// traits:
///
/// - `serde::Serialize`
/// - `serde::Deserialize`
/// - `exonum::encoding::Field`
///
/// This macro implements the [`ExonumJson`] trait. `ExonumJson`,
/// together with [`Field`], allows using data within persistent data structures in Exonum.
///
/// For additional information, refer to the [`encoding`] module documentation.
///
/// **Note.** The way this macros implements traits might not be optimal.
///
/// [`ExonumJson`]: ./encoding/serialize/json/trait.ExonumJson.html
/// [`Field`]: ./encoding/trait.Field.html
/// [`encoding`]: ./encoding/index.html
#[macro_export]
macro_rules! implement_exonum_serializer {
    ($name:ident) => {
        impl $crate::encoding::serialize::json::ExonumJsonDeserialize for $name {
            fn deserialize(
                value: &$crate::encoding::serialize::json::reexport::Value,
            ) -> Result<$name, Box<::std::error::Error>> {
                use $crate::encoding::serialize::json::reexport::from_value;
                Ok(from_value(value.clone())?)
            }
        }

        impl $crate::encoding::serialize::json::ExonumJson for $name {
            fn deserialize_field<B>(
                value: &$crate::encoding::serialize::json::reexport::Value,
                buffer: &mut B,
                from: $crate::encoding::Offset,
                to: $crate::encoding::Offset,
            ) -> Result<(), Box<::std::error::Error>>
            where
                B: $crate::encoding::serialize::WriteBufferWrapper,
            {
                use $crate::encoding::serialize::json::reexport::from_value;
                let value: $name = from_value(value.clone())?;
                buffer.write(from, to, value);
                Ok(())
            }

            fn serialize_field(
                &self,
            ) -> Result<
                $crate::encoding::serialize::json::reexport::Value,
                Box<::std::error::Error + Send + Sync>,
            > {
                use $crate::encoding::serialize::json::reexport::to_value;
                Ok(to_value(self)?)
            }
        }
    };
}

/// implement serializing wrappers and methods for json
#[macro_use]
pub mod json;

/// `WriteBufferWrapper` is a trait specific for writing fields in place.
#[doc(hidden)]
pub trait WriteBufferWrapper {
    fn write<'a, T: Field<'a>>(&'a mut self, from: Offset, to: Offset, val: T);
}

impl WriteBufferWrapper for MessageWriter {
    fn write<'a, T: Field<'a>>(&'a mut self, from: Offset, to: Offset, val: T) {
        self.write(val, from, to)
    }
}

impl WriteBufferWrapper for Vec<u8> {
    fn write<'a, T: Field<'a>>(&'a mut self, from: Offset, to: Offset, val: T) {
        val.write(self, from, to)
    }
}

/// Reexport of `serde` specific traits, this reexports
/// provide compatibility layer with important `serde` version.
pub mod reexport {
    pub use serde::de::{DeserializeOwned, Error as DeError};
    pub use serde::ser::Error as SerError;
    pub use serde::ser::SerializeStruct;
    pub use serde::{Deserialize, Deserializer, Serialize, Serializer};
}
